from running.benchmark import JavaBenchmark
from running.modifier import *
from running.config import Configuration


def test_jvm_arg():
    j = JVMArg(name="j", val="-Xms100M -D\"foo bar\"")
    assert j.val == ["-Xms100M", "-Dfoo bar"]


def test_jvm_classpath():
    j = JVMClasspath(name="j", val="/bin /foo \"/Users/John Citizen/\"")
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"],
        suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["fizzbuzz", "/bin", "/foo", "/Users/John Citizen/"]


def test_jvm_classpath_append():
    j = JVMClasspathAppend(name="j", val="/bin /foo \"/Users/John Citizen/\"")
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"],
        suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["fizzbuzz", "/bin", "/foo", "/Users/John Citizen/"]


def test_jvm_classpath_prepend():
    j = JVMClasspathPrepend(name="j", val="/bin /foo \"/Users/John Citizen/\"")
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"],
        suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["/bin", "/foo", "/Users/John Citizen/", "fizzbuzz"]


def test_program_arg():
    p = ProgramArg(name="p", val="/bin /foo \"/Users/John Citizen/\"")
    assert p.val == ["/bin", "/foo", "/Users/John Citizen/"]


def test_expand_value_opts():
    p = EnvVar(name="path", var="PATH", val="{0}:{1}")
    assert p.val == "{0}:{1}"
    p = p.apply_value_opts(value_opts=["/bin", "/sbin"])
    assert p.val == "/bin:/sbin"


def test_modifier_set():
    c = Configuration({
        "modifiers": {
            "a": {
                "type": "JVMArg",
                "val": "-XX:GC={0}"
            },
            "b": {
                "type": "EnvVar",
                "var": "FOO",
                "val": "BAR"
            },
            "c": {
                "type": "EnvVar",
                "var": "FIZZ",
                "val": "BUZZ"
            },
            "set": {
                "type": "ModifierSet",
                "val": "a-{0}|b"
            },
            "set_nested": {
                "type": "ModifierSet",
                "val": "set-{0}|c"
            }
        }
    })
    c.resolve_class()
    mods = c.get("modifiers")["set"].apply_value_opts(
        value_opts=["NoGC"]).flatten(c)
    mods = c.get("modifiers")["set_nested"].apply_value_opts(
        value_opts=["NoGC"]).flatten(c)
    assert len(mods) == 3



from running.modifier import JVMArg, Modifier, JSArg
from typing import Any, Dict, Union
from pathlib import Path
import logging
from running.util import register


class Runtime(object):
    CLS_MAPPING: Dict[str, Any]
    CLS_MAPPING = {}

    def __init__(self, name: str, **kwargs):
        self.name = name

    @staticmethod
    def from_config(name: str, config: Dict[str, str]) -> Any:
        return Runtime.CLS_MAPPING[config["type"]](name=name, **config)

    def get_executable(self) -> Union[str, Path]:
        raise NotImplementedError

    def get_heapsize_modifier(self, size: int) -> Modifier:
        raise NotImplementedError

    def is_oom(self, _output: bytes) -> bool:
        raise NotImplementedError


class DummyRuntime(Runtime):
    def __init__(self, executable: str):
        super().__init__(name="dummy")
        self.executable = executable

    def get_executable(self) -> Union[str, Path]:
        return self.executable

    def is_oom(self, _output: bytes) -> bool:
        return False


@register(Runtime)
class NativeExecutable(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_executable(self) -> Union[str, Path]:
        return ""

    def is_oom(self, _output: bytes) -> bool:
        return False


class JVM(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_executable(self) -> Path:
        raise NotImplementedError

    def __str__(self):
        return "JVM {}".format(self.name)

    def get_heapsize_modifier(self, size: int) -> Modifier:
        size_str = "{}M".format(size)
        heapsize = JVMArg(
            name="heap{}".format(size_str),
            val="-Xms{} -Xmx{}".format(size_str, size_str)
        )
        return heapsize

    def is_oom(self, output: bytes) -> bool:
        for pattern in [b"Allocation Failed", b"OutOfMemoryError", b"ran out of memory", b"panicked at 'Out of memory!'"]:
            if pattern in output:
                return True
        return False


@register(Runtime)
class OpenJDK(JVM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.release = kwargs["release"]
        try:
            self.release = int(self.release)
        except ValueError:
            raise TypeError("The release of an OpenJDK has to be int-like")
        self.home: Path
        self.home = Path(kwargs["home"])
        if not self.home.exists():
            logging.warning("OpenJDK home {} doesn't exist".format(self.home))
        self.executable = self.home / "bin" / "java"
        if not self.executable.exists():
            logging.warning(
                "{} not found in OpenJDK home".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable

    def __str__(self):
        return "{} OpenJDK {} {}".format(super().__str__(), self.release, self.home)


@register(Runtime)
class JikesRVM(JVM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.home: Path
        self.home = Path(kwargs["home"])
        if not self.home.exists():
            logging.warning("JikesRVM home {} doesn't exist".format(self.home))
        self.executable = self.home / "rvm"
        if not self.home.exists():
            logging.warning(
                "{} not found in JikesRVM home".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable

    def __str__(self):
        return "{} JikesRVM {}".format(super().__str__(), self.home)


class JavaScriptRuntime(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.executable: Path
        self.executable = Path(kwargs["executable"])
        if not self.executable.exists():
            logging.warning(
                "JavaScriptRuntime executable {} doesn't exist".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable


@register(Runtime)
class D8(JavaScriptRuntime):
    def __str__(self):
        return "{} d8 {}".format(super().__str__(), self.executable)

    def get_heapsize_modifier(self, size: int) -> Modifier:
        size_str = "{}".format(size)
        heapsize = JSArg(
            name="heap{}".format(size_str),
            val="--initial-heap-size={} --max-heap-size={}".format(
                size_str, size_str)
        )
        return heapsize

    def is_oom(self, output: bytes) -> bool:
        # The format is "Fatal javascript OOM in ..."
        # such as "Fatal javascript OOM in Reached heap limit"
        # or "Fatal javascript OOM in Ineffective mark-compacts near heap limit"
        return b"Fatal javascript OOM in" in output


@register(Runtime)
class SpiderMonkey(JavaScriptRuntime):
    def __str__(self):
        return "{} SpiderMonkey {}".format(super().__str__(), self.executable)

    def get_heapsize_modifier(self, size: int) -> Modifier:
        size_str = "{}".format(size)
        # FIXME doesn't seem to be working
        heapsize = JSArg(
            name="heap{}".format(size_str),
            val="--available-memory={}".format(size_str)
        )
        return heapsize

    def is_oom(self, output: bytes) -> bool:
        # FIXME not sure how to check for OOM for SpiderMonkey yet
        return False


@register(Runtime)
class JavaScriptCore(JavaScriptRuntime):
    def __str__(self):
        return "{} JavaScriptCore {}".format(super().__str__(), self.executable)

    def get_heapsize_modifier(self, size: int) -> Modifier:
        size_str = "{}".format(size)
        # FIXME doesn't seem to be working
        heapsize = JSArg(
            name="heap{}".format(size_str),
            val="--gcMaxHeapSize={}".format(size_str)
        )
        return heapsize

    def is_oom(self, output: bytes) -> bool:
        # FIXME not sure how to check for OOM for JavaScriptCore yet
        return False



from running.config import Configuration
from running.runtime import DummyRuntime
from running.suite import BinaryBenchmarkSuite


def test_binary_benchmark_suite_quoted():
    b = BinaryBenchmarkSuite(
        name="foobar",
        programs={
            "ls": {
                "path": "/bin/ls",
                "args": "foo \"bar buzz\""
            }
        }
    )
    assert b.programs["ls"]["args"] == ["foo", "bar buzz"]


def test_dacapo_size():
    c = Configuration({
        "suites": {
            "dacapo2006": {
                "type": "DaCapo",
                "release": "2006",
                "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                "timing_iteration": 3
            }
        },
        "benchmarks": {
            "dacapo2006": [
                "fop",
                dict(name="fop_small", bm_name="fop", size="small")
            ]
        }
    })

    c.resolve_class()
    fop = c.get("benchmarks")["dacapo2006"][0]
    fop_small = c.get("benchmarks")["dacapo2006"][1]
    assert "-s default" in fop.to_string(DummyRuntime("java"))
    assert "-s small" in fop_small.to_string(DummyRuntime("java"))


def test_dacapo_timing_iteration():
    c = Configuration({
        "suites": {
            "dacapo2006": {
                "type": "DaCapo",
                "release": "2006",
                "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                "timing_iteration": 3
            }
        },
        "benchmarks": {
            "dacapo2006": [
                "fop",
                dict(name="fop_converge", bm_name="fop",
                     timing_iteration="converge")
            ]
        }
    })

    c.resolve_class()
    fop = c.get("benchmarks")["dacapo2006"][0]
    fop_converge = c.get("benchmarks")["dacapo2006"][1]
    assert "-n 3" in fop.to_string(DummyRuntime("java"))
    assert "-converge" in fop_converge.to_string(DummyRuntime("java"))



from running.command.log_preprocessor import filter_stats, reduce_stats, sum_work_perf_event, calc_work_ipc, ratio_work_perf_event


def test_filter():
    assert filter_stats(lambda x: "foo" in x)({
        "foo1": 1,
        "2foo": 2,
        "bar3": 3,
        "4bar": 4
    }) == {"foo1": 1, "2foo": 2}


def test_reduce():
    assert reduce_stats(r"\d*foo\d*", "foo_sum", lambda x, y: x + y)({
        "foo1": 1,
        "2foo": 2,
        "bar3": 3,
        "4bar": 4
    })["foo_sum"] == 3


def test_sum_perf():
    assert sum_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")({
        "work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.total": 1,
        "work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.total": 2,
        "work.ScanUniverseRoots.PERF_COUNT_HW_CACHE_L1D:MISS.total": 3
    })["work.PERF_COUNT_HW_CPU_CYCLES.total"] == 3


def test_ratio_perf():
    stats = {
        "work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.total": 2,
        "work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.total": 3,
        "work.foo.PERF_COUNT_HW_CPU_CYCLES.total": 5
    }
    stats = sum_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")(stats)
    ratios = ratio_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")(stats)
    assert ratios["work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.2
    assert ratios["work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.3
    assert ratios["work.foo.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.5


def test_ipc():
    ipcs = calc_work_ipc({
        "work.foo.PERF_COUNT_HW_CPU_CYCLES.total": 20,
        "work.foo.PERF_COUNT_HW_INSTRUCTIONS.total": 4,
        "work.bar.PERF_COUNT_HW_CPU_CYCLES.total": 10,
        "work.bar.PERF_COUNT_HW_INSTRUCTIONS.total": 5,
    })
    print(ipcs)
    assert ipcs["work.foo.INSTRUCTIONS_PER_CYCLE.ratio"] == 0.2
    assert ipcs["work.bar.INSTRUCTIONS_PER_CYCLE.ratio"] == 0.5



from pathlib import Path
from running.util import smart_quote, split_quoted


def test_split_quoted():
    assert split_quoted("123 \"foo bar\"") == ["123", "foo bar"]


def test_smart_quote():
    assert smart_quote(Path("/bin")/"123 456") == "\"/bin/123 456\""
